use std::collections::HashMap;
use std::env::args;
use std::ffi::OsStr;
use std::io::{stderr, stdin, stdout, Read};
use std::path::PathBuf;
use std::process::Stdio;
use zip::ZipArchive;
use serde::{Deserialize, Serialize};
use tokio::fs::File;
use tokio::io::AsyncWriteExt;

#[derive(Debug, Clone, Serialize, Deserialize, Eq, PartialEq, Hash)]
enum Side {
    #[serde(rename = "client")]
    Client,
    #[serde(rename = "server")]
    Server,
}

#[derive(Debug, Clone, Serialize, Deserialize, Eq, PartialEq, Hash)]
enum Requirement {
    #[serde(rename = "required")]
    Required,
    #[serde(rename = "optional")]
    Optional,
}

#[derive(Debug, Clone, Serialize, Deserialize, Eq, PartialEq, Hash)]
enum Dependency {
    #[serde(rename = "minecraft")]
    Minecraft,
    #[serde(rename = "forge")]
    Forge,
    #[serde(rename = "neoforge")]
    NeoForge,
    #[serde(rename = "fabric-loader")]
    Fabric,
}

impl Dependency {
    async fn download(&self, target: impl Into<PathBuf>, version: String, mc_version: String) -> Result<(), Box<dyn std::error::Error>> {
        let url = match self {
            Dependency::NeoForge => format!("https://maven.neoforged.net/releases/net/neoforged/forge/{0}/forge-{0}-universal.jar", version),
            Dependency::Fabric => format!(
                "https://meta2.fabricmc.net/v2/versions/loader/{}/{}/{}/server/jar",
                mc_version,
                version,
                reqwest::get("https://meta2.fabricmc.net/v2/versions/installer")
                    .await?
                    .json::<Vec<FabricInstallerVersion>>()
                    .await?
                    .first()
                    .ok_or("No fabric installer version found")?
                    .version
            ),
            Dependency::Minecraft => {
                let version_map = reqwest::get("https://raw.githubusercontent.com/liebki/MinecraftServerForkDownloads/refs/heads/main/release_vanilla_downloads.json")
                    .await?
                    .json::<HashMap<String, String>>()
                    .await?;
                
                version_map.get(&mc_version)
                    .ok_or("No Minecraft version found")?
                    .to_string()
            },
            Dependency::Forge => format!("https://maven.minecraftforge.net/net/minecraftforge/forge/{1}-{0}/forge-{1}-{0}-installer.jar", version, mc_version)
        };
        
        let target = target.into();
        if target.exists() {
            eprintln!("File {} already exists, skipping download", target.display());
            return Ok(());
        }

        let response = reqwest::get(url.clone()).await?;
        let status = response.status();
        if status.is_success() {
            let content = response.bytes().await?;
            let mut file = File::create(&target).await?;
            tokio::io::copy(&mut content.as_ref(), &mut file).await?;
            return Ok(())
        }
        Err(format!("Failed to download file: {}", url).into())
    }
}

#[derive(Debug, Clone, Eq, PartialEq, Serialize, Deserialize)]
struct ConfigFile {
    path: String,
    hashes: HashMap<String, String>,
    env: HashMap<Side, Requirement>,
    downloads: Vec<String>,
    #[serde(rename = "fileSize")]
    file_size: u64,
}

impl ConfigFile {
    async fn download(&self, target: impl Into<PathBuf>) -> Result<(), Box<dyn std::error::Error>> {
        let target = target.into();
        let target_path = target.join(&self.path);
        let mut disabled_path = target_path.extension().unwrap_or(OsStr::new("jar")).to_os_string();
        disabled_path.push(".disabled");
        if target_path.exists() || target_path.with_extension(disabled_path).exists() {
            eprintln!("File {} already exists (or is disabled), skipping download", target_path.display());
            return Ok(());
        }
        
        if let Some(parent) = target_path.parent() {
            tokio::fs::create_dir_all(parent).await?;
        }

        for link in &self.downloads {
            let response = reqwest::get(link).await?;
            if response.status().is_success() {
                println!("Downloading file from {} -> `{}`", link, &target_path.display());
                let content = response.bytes().await?;
                let mut file = File::create(&target_path).await?;
                tokio::io::copy(&mut content.as_ref(), &mut file).await?;

                return Ok(())
            } else {
                eprintln!("Failed to download file from {}", link);
            }
        }

        Err(format!("Failed to download file: {}", self.path).into())
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
struct ModrinthConfig {
    game: String,
    #[serde(rename = "formatVersion")]
    format_version: u8,
    #[serde(rename = "versionId")]
    version_id: String,
    name: String,
    summary: String,
    files: Vec<ConfigFile>,
    dependencies: HashMap<Dependency, String>,
}

impl ModrinthConfig {
    fn is_modded(&self) -> bool {
        self.dependencies.len() > 1
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
struct FabricInstallerVersion {
    url: String,
    maven: String,
    version: String,
    stable: bool,
}

#[derive(Debug, Clone)]
struct GameSetupParams {
    dependency: Dependency,
    version: String,
    mc_version: String,
}

impl GameSetupParams {
    fn new(config: &ModrinthConfig) -> Self {
        let (server, server_version, mc_version) = if config.is_modded() {
            let (dep, version) = config.dependencies.iter().find(|(k, _)| **k != Dependency::Minecraft).expect("Modded dependency version not found");
            let mc_version = config.dependencies.iter().find(|(k, _)| **k == Dependency::Minecraft).map(|(_, v)| v.clone()).expect("Minecraft version not found");

            (dep.clone(), version.clone(), mc_version)
        } else {
            let (dep, version) = config.dependencies.iter().find(|(k, _)| **k == Dependency::Minecraft).expect("Modded dependency version not found");
            (dep.clone(), version.clone(), version.clone())
        };
        
        Self {
            dependency: server,
            version: server_version,
            mc_version,
        }
    }
}


async fn get_config_from_archive(archive: &mut ZipArchive<std::fs::File>, out_dir: impl Into<PathBuf>) -> Option<ModrinthConfig> {
    let overrides = "overrides/";
    let modrinth_index = "modrinth.index.json";
    let out_dir = out_dir.into();

    for i in 0..archive.len() {
        let mut file = archive.by_index(i).expect("Unable to read file");

        if file.name() == modrinth_index {
            let mut contents = String::new();
            if let Ok(_) = file.read_to_string(&mut contents) {
                return Some(serde_json::from_str::<ModrinthConfig>(contents.as_str()).expect("Unable to parse modrinth config"))
            } else {
                eprintln!("Can't read contents");
            }
        } else if file.name().starts_with(overrides) && file.is_file() {
            if let Some(path) = file.enclosed_name() {
                let realpath = path.parent().unwrap().to_str().unwrap();
                if realpath.len() >= overrides.len() {
                    let realpath = &realpath[overrides.len()..];
                    
                    // Create directory "out/" + realpath if it doesn't exist
                    let out_path = out_dir.join(realpath);
                    tokio::fs::create_dir_all(out_path.clone()).await.ok()?;

                    let mut out_file = std::fs::File::create(out_path.join(path.file_name().unwrap())).ok()?;
                    std::io::copy(&mut file, &mut out_file).expect("Unable to copy file");
                } else {
                    // Probably just a directory: SKIP
                    continue;
                }
            }
        } else {
            println!("File {} is not a modrinth index file", file.name());
        }
    }
    None
}

async fn get_config_from_file(file: &str) -> Option<ModrinthConfig> {
    let file = std::fs::File::open(file).expect("Unable to open file");
    let mut archive = ZipArchive::new(file).expect("Unable to read zip archive");
    get_config_from_archive(&mut archive, "out/").await
}

#[tokio::main]
async fn main() {
    let args: Vec<String> = args().collect();
    if args.len() < 2 || args.len() > 4 {
        eprintln!("Usage: {} <mrpack> [target dir] [previous version]", args[0]);
        return;
    }
    let out_dir = match args.len() {
        3 | 4 => PathBuf::from(&args[2]),
        _ => PathBuf::from("out/"),
    };
    let previous_version = match args.len() {
        4 => Some(get_config_from_file(args[3].as_str()).await.unwrap()),
        _ => None,
    };
    let file = std::fs::File::open(&args[1]).expect("Unable to open file");
    let mut archive = ZipArchive::new(file).expect("Unable to read zip archive");

    // Create directory "out" and subdirectory "mods"
    std::fs::create_dir_all(&out_dir).expect("Unable to create directory");
    
    let eula = "#By changing the setting below to TRUE you are indicating your agreement to our EULA (https://aka.ms/MinecraftEULA).
#Thu Jan 01 00:00:00 GMT 1970
eula=true
";
    let mut eula_file = File::create(out_dir.join("eula.txt")).await.unwrap();
    eula_file.write_all(eula.as_bytes()).await.unwrap();

    println!("Extracting files from {}", args[1]);
    let config = get_config_from_archive(&mut archive, &out_dir).await.unwrap();
    for entry in &config.files {
        entry.download(&out_dir).await.unwrap();
    }
    
    if let Some(ref previous_version) = previous_version {
        for removal in previous_version.files.iter().filter(|v| config.files.iter().filter(|v1| v1.path == v.path).count() == 0).map(|c| {
            tokio::fs::remove_file(c.path.clone())
        }) {
            let _ = removal.await;
        }
    }
    
    let old_params = previous_version.as_ref().map(|prev| GameSetupParams::new(prev));
    let params = GameSetupParams::new(&config);
    let server_jar_name = if matches!(params.dependency, Dependency::Forge) {
        "forge-installer.jar"
    } else {
        "server.jar"
    };
    let server_jar = out_dir.join(server_jar_name);

    if match old_params { None => true, Some(old_params) => old_params.dependency != params.dependency || !server_jar.exists() } {
        let _ = tokio::fs::remove_file(&server_jar).await;
        println!("Downloading server jar");
        params.dependency.download(&server_jar, params.version, params.mc_version).await.unwrap();
        if let Dependency::Forge = params.dependency {
            match java_locator::locate_java_home() {
                Ok(java_home) => {
                    println!("Found Java at {}. Attempting to run installer...", java_home);
                    let java_home = <String as Into<PathBuf>>::into(java_home).join("bin").join(format!("java{}", std::env::consts::EXE_SUFFIX));
                    println!("Java home: {}", java_home.display());
                    let result = std::process::Command::new(&java_home)
                        .current_dir(out_dir)
                        .arg("-jar")
                        .arg(server_jar_name)
                        .arg("--installServer")
                        .stderr(Stdio::inherit())
                        .stdout(Stdio::inherit())
                        .stdin(Stdio::inherit())
                        .spawn();
                    match result {
                        Ok(mut child) => {
                            let _ = child.wait();
                            println!("Installer finished");
                            let _ = tokio::fs::remove_file(&server_jar).await;
                        }
                        Err(e) => {
                            eprintln!("Error running installer: {}", e);
                        }
                    }
                }
                Err(e) => {
                    eprintln!("Error locating Java! You will need to execute the Forge installer manually: `java -jar {server_jar_name} --installServer`");
                }
            }
        }
    } else {
        println!("Server jar already exists, skipping download");
    }
}
